package edu.wpi.aipaint.ai;

import java.util.ArrayList;
import java.util.Stack;

import edu.wpi.aipaint.directo.BTAIController;
import edu.wpi.aipaint.directo.DirectoProgram;
import edu.wpi.aipaint.main.MainWindow;
import edu.wpi.aipaint.pipeline.AIProcessor;
import edu.wpi.aipaint.pipeline.Direction;
import edu.wpi.aipaint.pipeline.PipelineModule;

/**
 * Performs modifications to an AI data representations according to given
 * Directions.
 */
public class BTAIModule implements AIProcessor {
	private PipelineModule pipeline;

	// Each input List of Directions is stored in this directionUndoStack,
	// while a List of complementary Directions is stored in this
	// complementUndoStack. Both the Directions and their complements
	// are stored so that we don't have to recompute the complements on each
	// undo or redo action.
	// The redo Stacks store Lists that have been popped off the undo Stacks
	// until more input is received from the user.
	private Stack<DirectoProgram> directionUndoStack = new Stack<DirectoProgram>();
	private Stack<DirectoProgram> directionRedoStack = new Stack<DirectoProgram>();
	private Stack<DirectoProgram> complementUndoStack = new Stack<DirectoProgram>();
	private Stack<DirectoProgram> complementRedoStack = new Stack<DirectoProgram>();

	private final String REPLACE_AS_ROOT = "Make priority";
	private final String ADD_AT_END = "Add at end";
	private final String DO_NOT_ADD = "Do not add";

	public BTAIModule() {
		directionUndoStack = new Stack<DirectoProgram>();
		directionRedoStack = new Stack<DirectoProgram>();
		complementUndoStack = new Stack<DirectoProgram>();
		complementRedoStack = new Stack<DirectoProgram>();
	}

	@Override
	public void applyDirectioProgram(DirectoProgram directions) {

		BTAIController aiData = (BTAIController) pipeline.getAIAgent()
				.getAIController();
		// Execute the Directions and store their complements.
		DirectoProgram complements = new DirectoProgram(directions.getWorldState());

		// Construct the behavior tree
		BehaviorTree bt = new BehaviorTree();

		for (Direction direction : directions.getDirections()) {
			try {
				direction.execute(bt.getRoot(), directions.getWorldState());
			} catch (DiscourseException e) {
				// Fresh behavior tree, shouldn't happen
				System.err.println(e.getLocalizedMessage());
				e.printStackTrace();
			}
		}

		try {
			aiData.getBehaviorTree().mergeBT(bt);
		} catch (BTMergeException e) {
			try {
				bt.mergeBT(aiData.getBehaviorTree());
			} catch (BTMergeException e1) {
				// Trigger discourse here
				ArrayList<String> answers = new ArrayList<String>();
				answers.add("Nope");
				answers.add(REPLACE_AS_ROOT);
				answers.add(ADD_AT_END);
				answers.add(DO_NOT_ADD);
				try {
					this.pipeline.getMainWindow()
							.askMultipleChoiceQuestion("Whoopsie, could not merge",
									answers);
				} catch (Exception e2) {
					// TODO Auto-generated catch block
					e2.printStackTrace();
				}
			}
		}

		// Add the Direction List to the directionUndoStack
		// and the complement List to the complementUndoStack.
		directionUndoStack.push(directions);
		complementUndoStack.push(complements);

		// Empty the directionRedoStack and the complementRedoStack.
		directionRedoStack.clear();
		complementRedoStack.clear();

		// Move on to the next step in the pipeline.
		pipeline.aiProcessorDone();
	}

	@Override
	public void undo() {
		BTAIController aiData = (BTAIController) pipeline.getAIAgent()
				.getAIController();

		if (!complementUndoStack.empty() && !directionUndoStack.empty()) {
			DirectoProgram complements = complementUndoStack.pop();
			complementRedoStack.push(complements);
			directionRedoStack.push(directionUndoStack.pop());

			for (Direction complement : complements.getDirections()) {
				// Apply the complement Direction to the AI data representation.
				try {
					complement.execute(aiData.getBehaviorTree().getRoot(),
							complements.getWorldState());
				} catch (DiscourseException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
		}
	}

	@Override
	public void redo() {
		BTAIController aiData = (BTAIController) pipeline.getAIAgent()
				.getAIController();

		if (!complementRedoStack.empty() && !directionRedoStack.empty()) {
			DirectoProgram directions = directionRedoStack.pop();
			directionUndoStack.push(directions);
			complementUndoStack.push(complementUndoStack.pop());

			for (Direction direction : directions.getDirections()) {
				// Apply the Direction to the AI data representation.
				try {
					direction.execute(aiData.getBehaviorTree().getRoot(),
							directions.getWorldState());
				} catch (DiscourseException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
		}
	}

	@Override
	public void activate(MainWindow window) {
	}

	@Override
	public void deactivate() {
	}

	@Override
	public String getName() {
		return "Behavior Tree AI Module";
	}

	@Override
	public void tick() {
	}

	@Override
	public void setPipelineParent(PipelineModule parent) {
		this.pipeline = parent;
	}
}
