package edu.wpi.aipaint.ai;

import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.TreeSet;

import edu.wpi.aipaint.attributes.Attribute;
import edu.wpi.aipaint.main.WorldState;
import edu.wpi.aipaint.pipeline.Action;
import edu.wpi.aipaint.values.ContinuousValue;
import edu.wpi.aipaint.values.FloatWrapper;
import edu.wpi.aipaint.values.Value;

/**
 * A BTNode for handling continuous decisions.
 * 
 * @author Jesse Bassett
 */
public class BTContinuousDecisionNode implements IBTDecisionNode {
	protected Attribute attribute;
	protected IBTDecisionNode parent;
	protected List<ContinuousValue> decisionSlots;
	protected HashMap<ContinuousValue, IBTNode> children;
	protected Value minValue, maxValue;

	public BTContinuousDecisionNode(
			Attribute attribute, IBTDecisionNode parent,
			List<ContinuousValue> decisionSlots, FloatWrapper minValue,
			FloatWrapper maxValue) {
		this.attribute = attribute;
		this.parent = parent;
		this.decisionSlots = decisionSlots;
		this.children = new HashMap<ContinuousValue, IBTNode>(decisionSlots.size());
		this.minValue = minValue;
		this.maxValue = maxValue;
	}

	public IBTDecisionNode attachNextChild(	IBTNode p_child,
											WorldState p_ws)
			throws DiscourseException {
		if (!decisionSlots.isEmpty()) {
			ContinuousValue val = decisionSlots.remove(0);

			this.children.put(val, p_child);

			return this;
		} else {
			IBTNode child = this.children.get(this.attribute.evaluate(p_ws));
			if (child instanceof IBTLeafNode) {
				throw new DiscourseException("No free spots found!",
						this,
						DiscourseException.TYPE.NO_AVAILABLE_MERGE);
			} else {
				return ((IBTDecisionNode) child).attachNextChild(p_child, p_ws);
			}
		}
	}

	@Override
	public Action evaluate(WorldState p_ws) {
		IBTNode child = this.getChild(p_ws);
		if (child instanceof IBTDecisionNode) {
			return ((IBTDecisionNode) child).evaluate(p_ws);
		} else if (child instanceof IBTLeafNode) {
			return ((IBTLeafNode) child).getAction();
		} else {
			System.err.println("There is no valid move");
			return null;
		}
	}

	@Override
	public Attribute getAttribute() {
		return this.attribute;
	}

	@Override
	public IBTNode getChild(Value p_attributeValue) {
		ContinuousValue cv = (ContinuousValue) p_attributeValue;
		TreeSet<ContinuousValue> ts = new TreeSet<ContinuousValue>();
		ts.addAll(children.keySet());
		return this.children.get(ts.ceiling(cv));
	}

	@Override
	public IBTNode getChild(WorldState p_ws) {
		return this.getChild(attribute.evaluate(p_ws));
	}

	@Override
	public int getChildCount() {
		return this.children.size();
	}

	@Override
	public Iterator<? extends IBTNode> getChildrenIterator() {
		return this.children.values().iterator();
	}

	@Override
	public List<? extends Value> getDecisionSlots() {
		return this.decisionSlots;
	}

	/**
	 * @param p_child
	 *            {@link BTNode} to find the lower bound of
	 * @return The {@link Value} of the lower bound for this decision
	 */
	public Value getLowerBound(IBTNode p_child) {
		Collections.sort(this.decisionSlots);
		for (int i = 0; i < this.decisionSlots.size() - 1; i++) {
			if (this.children.get(this.decisionSlots.get(i + 1)) == p_child) {
				return this.decisionSlots.get(i);
			}
		}
		return null;
	}

	@Override
	public IBTDecisionNode getParent() {
		return this.parent;
	}

	/**
	 * @param p_child
	 *            {@link BTNode} to find the upper bound of
	 * @return The {@link Value} of the upper bound
	 */
	public Value getUpperBound(IBTNode p_child) {
		for (int i = 0; i < this.decisionSlots.size(); i++) {
			if (this.children.get(this.decisionSlots.get(i)) == p_child) {
				return this.decisionSlots.get(i);
			}
		}
		return null;
	}

	@Override
	public Value getValue(IBTNode child) {
		for (Value k : this.decisionSlots) {
			if (child.equals(this.children.get(k)))
				return k;
		}
		return null;
	}

	@Override
	public IBTDecisionNode insertChild(	Value p_value,
										IBTNode p_child) {
		ContinuousValue cv;
		if (p_value instanceof ContinuousValue) {
			cv = (ContinuousValue) p_value;
			this.decisionSlots.add(cv);
			this.children.put(cv, p_child);
			return this;
		} else {
			System.err.println("Invalid type of value.");
			return null;
		}
	}

	@Override
	public void mergeChildren(IBTDecisionNode p_other)
			throws DiscourseException {
		if (this.attribute.equals(p_other.getAttribute())) {
			for (Value v : p_other.getDecisionSlots()) {
				if (this.children.containsKey(v))
					throw new DiscourseException("Value " + v
							+ " already exists.",
							this,
							DiscourseException.TYPE.DUPLICATE_NODE);
				this.decisionSlots.add((ContinuousValue) v);
				this.children.put((ContinuousValue) v, p_other.getChild(v));
			}
		}
	}

	@Override
	public IBTNode removeChild(IBTNode p_child) {
		if (this.children.containsValue(p_child)) {
			for (Value k : this.children.keySet()) {
				if (this.children.get(k) == p_child) {
					this.decisionSlots.remove(k);
					IBTNode child = this.children.get(k);
					this.children.remove(k);
					return child;
				}
			}
			/* ffffffffuuuuuuuuuuuuu */
			return null;
		} else {
			return null;
		}
	}

	@Override
	public IBTNode removeChild(Value p_value) {
		if (this.children.containsKey(p_value)) {
			this.decisionSlots.remove(p_value);
			IBTNode child = this.children.get(p_value);
			this.children.remove(p_value);
			return child;
		} else {
			return null;
		}
	}

	@Override
	public void setParent(IBTDecisionNode parent) {
		this.parent = parent;
	}

	public String toXML() {
		throw new UnsupportedOperationException();
	}

	public String treePrint(int p_depth) {
		String ret = ""; 
		TreeSet<ContinuousValue> ts = new TreeSet<ContinuousValue>();
		ts.addAll(this.children.keySet());
		Iterator<ContinuousValue> tsi = ts.iterator();
		while (tsi.hasNext()) {
			ContinuousValue cv = tsi.next();
			if(!(ts.first().equals(cv)))
				ret += "ELSE ";
			ret += "IF " + this.attribute + " <= ";
			for (int i = 0; i < p_depth; i++) {
				ret += "| ";
			}
			if (ts.last().equals(cv)) {
				ret += "INF";
			} else {
				ret += cv.toString();
			}

			ret += " do\n";

			IBTNode child = this.children.get(cv);

			ret += child.treePrint(p_depth + 1);
		}
		return ret;
	}
}