package edu.wpi.aipaint.ai;

import java.util.ArrayList;

import edu.wpi.aipaint.main.GameKeeper;
import edu.wpi.aipaint.main.WorldState;
import edu.wpi.aipaint.pipeline.Action;

public class BehaviorTree {
	/**
	 * Root node for the tree
	 */
	private IBTDecisionNode m_root;

	public BehaviorTree(Action action) {
		this.m_root = new BTDefaultRootNode(action, this);
	}
	
	public BehaviorTree() {
		this.m_root = new BTDefaultRootNode(this);
	}

	public BehaviorTree(IBTDecisionNode root) {
		this.m_root = root;
	}

	/**
	 * Removes all elements from the tree.
	 */
	public void clearTree() {
		this.m_root = null;
	}

	/**
	 * Gets the root node of the tree
	 * 
	 * @return The root node
	 */
	public IBTNode getRoot() {
		return m_root;
	}

	/**
	 * Manually sets the root node of the tree. Can invalidate the tree if the
	 * root has no children.
	 * 
	 * @param root
	 *            IBTNode to be placed as root.
	 */
	public void setRoot(IBTDecisionNode root) {
		this.m_root = root;
	}

	/**
	 * Returns the active path of the behavior tree. Includes the leaf node at
	 * the end of the path.
	 * 
	 * @param ws
	 *            Current state of the world
	 * @return An array with the current active path of the AI
	 */
	public IBTNode[] getActivePath(WorldState ws) {
		IBTNode[] ret;
		ArrayList<IBTNode> activePath = new ArrayList<IBTNode>();
		IBTNode currentNode = this.m_root;
		activePath.add(this.m_root);

		while (currentNode != null && currentNode instanceof IBTDecisionNode) {
			IBTNode child = ((IBTDecisionNode) currentNode).getChild(ws);
			activePath.add(child);
			currentNode = child;
		}

		ret = new IBTNode[activePath.size()];
		activePath.toArray(ret);
		return ret;
	}

	/**
	 * @return Returns the tree in ASCII format
	 */
	public String printTree() {
		if (this.m_root == null) {
			return "Tree is empty.";
		}
		return this.m_root.treePrint(0);
	}

	/**
	 * @param ws
	 *            World state to evaluate for
	 * @return Action to perform
	 */
	public IBTNode getActiveNode(WorldState ws) {
		if (this.m_root == null) {
			return null;
		}
		IBTNode active = this.m_root;
		while (active != null && active instanceof IBTDecisionNode) {
			active = ((IBTDecisionNode) active).getChild(ws);
		}
		return active;
	}
	
	/**
	 * Merges another behavior tree into this one.
	 * @param bt Other {@link BehaviorTree} to merge into this one
	 * @return The root node
	 * @throws BTMergeException if there is an error.
	 */
	public IBTDecisionNode mergeBT(BehaviorTree bt) throws BTMergeException{
		try {
			this.m_root.attachNextChild(bt.getRoot(), GameKeeper.GetInstance().getCurrentWorldState());
		} catch (DiscourseException e) {
			throw new BTMergeException("Failed to merge", BTMergeException.MERGE_ERROR_TYPE.NON_TRIVIAL_MERGE);
		}
		
		return this.m_root;
	}
	
	public String toString() {
		return this.printTree();
	}
}
