package edu.wpi.aipaint.directo;

import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.Stack;

import edu.wpi.aipaint.main.WorldState;
import edu.wpi.aipaint.pipeline.Direction;

/**
 * A wrapper for a {@link Stack} of {@link Direction}s, with an accompanying
 * {@link WorldState}. This helps to ensure that the program will run
 * idempotently. If build from a tree, assumes that the nodes are inserted
 * post-order, and the root node is inserted last.
 * 
 * @author Jesse Bassett
 */
public class DirectoProgram {
	/**
	 * Directions to execute in order.
	 */
	private Queue<Direction> directions;

	/**
	 * State to execute in.
	 */
	private WorldState ws;

	/**
	 * The list of directions is created empty.
	 * 
	 * @param ws
	 *            {@link WorldState} to identify by
	 */
	public DirectoProgram(WorldState ws) {
		super();
		this.directions = new LinkedList<Direction>();
		this.ws = ws;
	}

	/**
	 * Default constructor with a list of {@link Direction}s and a
	 * {@link WorldState}
	 */
	@SuppressWarnings("unchecked")
	public DirectoProgram(List<Direction> directoInput, WorldState ws) {
		super();
		this.directions = (Queue<Direction>) directoInput;
		this.ws = ws;
	}

	/**
	 * @param direction
	 *            {@link Direction} to add to the queue.
	 */
	public void addDirection(Direction direction) {
		this.directions.add(direction);
	}

	/**
	 * @return The {@link Direction} at the head.
	 */
	public Direction getDirection() {
		return this.directions.peek();
	}

	/**
	 * @return The {@link Direction} at the head of the queue, and removes it.
	 */
	public Direction retrieveDirection() {
		return this.directions.poll();
	}

	/**
	 * @return The {@link WorldState} conditions to add at.
	 */
	public WorldState getWorldState() {
		return this.ws;
	}

	/**
	 * @return A compliment {@link DirectoProgram} that will undo the effects of
	 *         the directions this will apply.
	 */
	public DirectoProgram generateCompliment() {
		DirectoProgram comp = new DirectoProgram(this.ws);
		Queue<Direction> tmp = this.directions;
		while (!tmp.isEmpty()) {
			comp.addDirection(tmp.poll().generateComplement());
		}

		return comp;
	}

	/**
	 * @return True if there are directions remaining
	 */
	public boolean hasDirections() {
		return !this.directions.isEmpty();
	}

	/**
	 * @return number of directions left to execute.
	 */
	public int size() {
		return this.directions.size();
	}

	public void addAll(List<Direction> directionsQueue) {
		this.directions.addAll(directionsQueue);
	}

	@SuppressWarnings("unchecked")
	public List<Direction> getDirections() {
		return (List<Direction>) this.directions;
	}
}