package edu.wpi.aipaint.main;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.KeyListener;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JPanel;

import edu.wpi.aipaint.pipeline.AIAgent;
import edu.wpi.aipaint.util.Line;
import edu.wpi.aipaint.util.Vector2;

@SuppressWarnings("serial")
public class MainWindow extends JFrame implements IQuestion, ComponentListener {

	private Game game;
	private Ticker ticker;

	private ArrayList<Module> modules;
	private Module activeModule;

	private JPanel sideBar;
	
	private InputListenerProxy inputListenerProxy;

	private IQuestion dialogueHandler;
	
	private boolean aiPaintPaused;

	public MainWindow(Game game, ArrayList<Module> modules) {
		this.game = game;
		GameKeeper.GetInstance().setGame(game);
		this.modules = modules;
		this.addComponentListener(this);
	}
	
	public void setModuleGlassPane(JComponent glassPane)
	{
		this.setGlassPane(glassPane);
		this.getGlassPane().setVisible(true);
		inputListenerProxy.setGlassPane(this.getGlassPane());
		this.getGlassPane().addMouseListener(inputListenerProxy);
		this.getGlassPane().addMouseMotionListener(inputListenerProxy);
		this.getGlassPane().validate();
		setSubComponentsSize();
	}
	
	public void setModuleMouseListener(MouseListener mouseListener)
	{
		inputListenerProxy.setMouseListener(mouseListener);
	}
	
	
	public void setModuleMouseMotionListener(MouseMotionListener mouseMotionListener)
	{
		inputListenerProxy.setMouseMotionListener(mouseMotionListener);
	}
	
	public void setModuleKeyListener(KeyListener keyListener)
	{
		inputListenerProxy.setKeyListener(keyListener);
	}

	private void activateModule(Module activeModule) {
		activeModule.activate(this);
		this.activeModule = activeModule;
	}

	public Module getModule(String name) {
		for (Module mm : modules) {
			if (mm.getName() == name) {
				return mm;
			}
		}
		throw new RuntimeException("Main window does not have desired module: " + name);
	}

	@Override
	public String askMultipleChoiceQuestion(String question,
			List<String> answers) throws Exception {
		return dialogueHandler.askMultipleChoiceQuestion(question, answers);
	}

	public void draw() {
		game.draw(1);

		this.repaint();
	}

	public Game getGame() {
		return game;
	}

	public Poseable getPoseableAt(Vector2 worldSpacePoint) {
		return game.getPoseableAt(worldSpacePoint);
	}
	
	public List<Poseable> getPoseablesInRange(Vector2 worldSpaceCenter, float worldSpaceRadius) {
		return game.getPoseablesCollidingWith(worldSpaceCenter, worldSpaceRadius);
	}
	
	public AIAgent getAIAgentAt(Vector2 worldSpacePoint) {
		return game.getAIAgentAt(worldSpacePoint);
	}

	public Vector2 getScreenSpacePoint(Vector2 worldSpacePoint) {
		Vector2 screenSpacePoint = game.getScreenSpacePoint(worldSpacePoint);
		return new Vector2(screenSpacePoint.getX(),
				screenSpacePoint.getY());

	}

	public JPanel getSideBar() {
		return sideBar;
	}

	public Ticker getTicker() {
		return ticker;
	}

	public Vector2 getWorldSpacePoint(Vector2 screenSpacePoint) {
		return game.getWorldSpacePoint(new Vector2(screenSpacePoint.getX(),
				screenSpacePoint.getY()));
	}
	
	public void toggleAIPaintPause()
	{
		aiPaintPaused = !aiPaintPaused;
	}
	
	public boolean isAIPaintPaused()
	{
		return aiPaintPaused;
	}

	public void init() {
		game.aiPaintInit();

		aiPaintPaused = false;

		add((JPanel) game);
		
		sideBar = new JPanel();
		sideBar.setPreferredSize(new Dimension(150, 400));

		add(sideBar, BorderLayout.EAST);
		
		this.inputListenerProxy = new InputListenerProxy(this, sideBar, 
				this.getContentPane(), this.getGlassPane());
		inputListenerProxy.setGameMouseListener(game.getMouseListener());
		inputListenerProxy.setGameMouseMotionListener(game.getMouseMotionListener());
		inputListenerProxy.setGameKeyListener(game.getKeyboardListener());
		this.getGlassPane().addMouseListener(inputListenerProxy);
		this.getGlassPane().addMouseMotionListener(inputListenerProxy);
		addKeyListener(inputListenerProxy);
		
		try {
			activateModule(getModule("Pipeline"));
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	public void setDialogueHandler(IQuestion newHandler) {
		this.dialogueHandler = newHandler;
	}

	public void start(int framerate) {
		if (ticker == null) {
			ticker = new Ticker(this, framerate);
			ticker.start();
		}
	}

	public void tick(int elapsedTime) {
		if (!aiPaintPaused)
			game.tick(elapsedTime);
		
		activeModule.tick();
		this.requestFocus();
		setSubComponentsSize();
	}

	@Override
	public void componentHidden(ComponentEvent arg0) { }

	@Override
	public void componentMoved(ComponentEvent arg0) { }

	@Override
	public void componentResized(ComponentEvent arg0) {
		setSubComponentsSize();
	}
	
	private void setSubComponentsSize() {
		Component c = (Component)game;
		Dimension d = getSize();
		d.setSize(d.getWidth() - sideBar.getSize().getWidth() - 10, d.getHeight() - 10);	//subtract 10 because it works better o_O
		c.setBounds(0, 0, (int)d.getWidth(), (int)d.getHeight());
		c.setSize(d);
	}

	@Override
	public void componentShown(ComponentEvent arg0) { }
}

class MainWindowGlassPane extends JComponent {
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	Line[] lines;

	protected void paintComponent(Graphics g) {
		for (Line l : lines) {
			g.setColor(Color.WHITE);
			g.drawLine((int) l.getPoint1().getX(), (int) l.getPoint1().getY(),
					(int) l.getPoint2().getX(), (int) l.getPoint2().getY());
		}
	}

	protected void setLines(Line[] newLines) {
		lines = newLines;
	}

}
