package edu.wpi.aipaint.pipeline;

import java.util.List;

import edu.wpi.aipaint.directo.DirectoProgram;
import edu.wpi.aipaint.main.GameKeeper;
import edu.wpi.aipaint.main.MainWindow;
import edu.wpi.aipaint.main.Module;
import edu.wpi.aipaint.util.TreeNode;
import edu.wpi.aipaint.util.UserObjectTreeNode;

/**
 * Module that manages the translation of data from user input to AI,
 * via submodules RawInputProcessor, StatementProcessor, and AIProcessor.
 */
public class PipelineModule implements Module {
	PipelineSubModule agentSelector;
	RawInputProcessor rawInputProcessor;
	StatementProcessor statementProcessor;
	AIProcessor aiProcessor;
	
	AIAgent activeAgent;
	
	TreeNode statementInput;
	List<Direction> directoInput;
	
	Module activeModule;
	
	MainWindow window;
	
	public PipelineModule(PipelineSubModule agentSelector, RawInputProcessor rawInputProcessor, 
			StatementProcessor statementProcessor, AIProcessor aiProcessor)
	{
		this.agentSelector = agentSelector;
		this.rawInputProcessor = rawInputProcessor;
		this.statementProcessor = statementProcessor;
		this.aiProcessor = aiProcessor;
		
		this.agentSelector.setPipelineParent(this);
		this.rawInputProcessor.setPipelineParent(this);
		this.statementProcessor.setPipelineParent(this);
		this.aiProcessor.setPipelineParent(this);
		
		directoInput = null;
	}
	
	/**
	 * Sets the AI under construction in this pipeline to the given AI.
	 */
	public void setAIAgent(AIAgent selectedAIAgent) {
		this.activeAgent = selectedAIAgent;
		setActiveModule(rawInputProcessor);
	}
	
	public void activateAgentSelection()
	{
		setActiveModule(agentSelector);
	}
	
	/**
	 * @return The currently active {@link MainWindow}
	 */
	public MainWindow getMainWindow() {
		return this.window;
	}
	
	/**
	 * Ends the raw input gathering phase of the pipeline, taking
	 * the given tree of Statements as the result.  This tree of
	 * Statements is used in the statement processing phase to
	 * produce a List of Directions.
	 * @param statementTree
	 */
	public void rawInputGathered(UserObjectTreeNode<Statement> statementTree)
	{
		this.statementInput = statementTree;
		statementProcessor.processStatements(statementTree);
	}
	
	/**
	 * Ends the statement processing phase of the pipeline, taking the given
	 * String of Directions in the Directo language as the result.
	 * The next phase is AI processing.
	 * @param serializedDirectoInput
	 */
	public void statementProcessingDone(String serializedDirectoInput)
	{
		// Convert the String of serialized Directions to a List<Direction>
		// and store it in directoInput before moving to the AI processing
		// phase.
	}
	
	/**
	 * Ends the statement processing phase of the pipeline, taking the given
	 * List of Directions as the result.
	 * The next phase is AI processing.
	 * @param serializedDirectoInput
	 */
	public void statementProcessingDone(List<Direction> directoInput)
	{
		this.directoInput = directoInput;
		DirectoProgram program = new DirectoProgram(directoInput, GameKeeper.GetInstance().getCurrentWorldState());
		// Move on to the AI processing phase.
		aiProcessor.applyDirectioProgram(program);
	}
	
	/**
	 * Ends the statement processing phase of the pipeline, taking the given
	 * List of Directions as the result.
	 * The next phase is AI processing.
	 * @param serializedDirectoInput
	 */
	public void statementProcessingDone(DirectoProgram program)
	{
		this.directoInput = program.getDirections();
		aiProcessor.applyDirectioProgram(program);
	}
	
	/**
	 * Ends the AI processing phase of the pipeline.  The AI data 
	 * representation will have been modified at this point.  
	 * The pipeline then resets to raw input processing.
	 */
	public void aiProcessorDone()
	{
		setActiveModule(rawInputProcessor);
	}
	
	public AIAgent getAIAgent()
	{
		return activeAgent;
	}
	
	/**
	 * Sets the active submodule to the given Module.
	 * @param nextModule
	 */
	private void setActiveModule(Module nextModule)
	{
		window.getSideBar().removeAll();
		activeModule = nextModule;
		nextModule.activate(window);
		window.setVisible(true);
	}
	
	@Override
	public void activate(MainWindow window) { 
		this.window = window;
		setActiveModule(agentSelector);
	}

	@Override
	public String getName() {
		return "Pipeline";
	}

	@Override
	public void tick() { 
		activeModule.tick();
	}

	@Override
	public void deactivate() { }
}
