package edu.wpi.aipaint.sketcho;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JButton;
import javax.swing.JToolBar;
import javax.swing.SwingUtilities;

import edu.wpi.aipaint.ai.IBTLeafNode;
import edu.wpi.aipaint.ai.IBTNode;
import edu.wpi.aipaint.attributes.Attribute;
import edu.wpi.aipaint.main.MainWindow;
import edu.wpi.aipaint.main.Poseable;
import edu.wpi.aipaint.main.WorldState;
import edu.wpi.aipaint.pipeline.AIController;
import edu.wpi.aipaint.pipeline.Action;
import edu.wpi.aipaint.pipeline.PipelineModule;
import edu.wpi.aipaint.pipeline.RawInputProcessor;
import edu.wpi.aipaint.pipeline.Statement;
import edu.wpi.aipaint.poseo.PoseModeMouseListener;
import edu.wpi.aipaint.shapedetector.Point;
import edu.wpi.aipaint.shapedetector.Recognizer;
import edu.wpi.aipaint.shapedetector.Result;
import edu.wpi.aipaint.util.DefaultTreeNode;
import edu.wpi.aipaint.util.Line;
import edu.wpi.aipaint.util.Line.LineFonts;
import edu.wpi.aipaint.util.TreeNode;
import edu.wpi.aipaint.util.Vector2;
import edu.wpi.aipaint.worldspacepoint.WorldSpacePoint;
import edu.wpi.aipaint.worldspacepoint.WorldSpacePointAttribute;
import edu.wpi.aipaint.worldspacepoint.WorldSpacePointPoseable;
import edu.wpi.aipaint.worldspacepoint.WorldSpacePointSketchListener;
import edu.wpi.aipaint.worldspacepoint.WorldSpacePointVector;

/**
 * Submodule of the Pipeline module that gathers raw input in the form
 * of Sketcho Statements written via a pick-and-place interface.
 */
public class SketchoModule implements RawInputProcessor, ActionListener, MouseListener, MouseMotionListener, KeyListener {
	
	private PipelineModule parent;
	private MainWindow window;
	
	private float minDistanceToSelect;
	private final float SCREEN_SPACE_MIN_SELECTION_DISTANCE = 10;
	
	private final float MINIMUM_RECOGNITION_THRESHOLD = .6f;
	private final float FULL_CONFIDENCE_THRESHOLD = .75f;
	private final float MULTI_RESULT_AMBIGUITY_THRESHOLD = .15f;
	private final long AMBIGUOUS_RECOGNITION_STALL_TIME = 1000;
	private final long MORE_INPUT_WAIT_TIME = 1750;
	private final float CLEANUP_CONTROL_POINT_THRESHOLD = 20;
	
	private float worldspaceCleanupThreshold;
	
	private DefaultTreeNode<SketchoStatement> activeStatement;
	
	private SketchoGlassPane pane;
	
	private ArrayList<SketchoStatement> renderedStatements;
	private boolean renderingStatements = false;
	
	private SketchoStatement selectedStatement;
	
	private StatementFactory factory;
	
	private long recognitionTime;
	
	private DefaultTreeNode<SketchoStatement> rootStatement;
	private DefaultTreeNode<SketchoStatement> currentParent;
	private boolean needsMoreStatements;
	
	private ArrayList<Point> currentStroke;	//The recognizer exists in float space, so we need to define them by float-defined points rather than awt's int defined points
	private ArrayList<ArrayList<Point>> strokes;
	
	private Vector2 lastClick;
	
	private Generalizer generalizer;
	
	private Recognizer recognizer;
	
	private PoseModeMouseListener poseListener;
	private int buttonDown = 0;
	
	
	private JButton showMeButton;
	
	private ArrayList<SketchoStatement> intermediateStatements;
	
	public SketchoModule(StatementFactory factory, HeuristicFactory heuristicFactory, Recognizer recognizer) {
		this.factory = factory;
		
		generalizer = new Generalizer(heuristicFactory);
		pane = new SketchoGlassPane();
		renderedStatements = new ArrayList<SketchoStatement>();
		intermediateStatements = new ArrayList<SketchoStatement>();
		currentStroke = new ArrayList<Point>();
		this.recognizer = recognizer;
	}
	
	@Override
	public void setPipelineParent(PipelineModule parent) {
		this.parent = parent;
		
	}

	@Override
	public void activate(MainWindow window) {
		strokes = new ArrayList<ArrayList<Point>>();
		recognitionTime = -1;
		this.window = window;
		window.getSideBar().add(makeToolBar(), BorderLayout.NORTH);
		
		needsMoreStatements = true;
		// Calibrate the mindistance with the current world space to screen space conversion.
		// Note that this value could be different every time we pause the game
		minDistanceToSelect = window.getWorldSpacePoint(new Vector2(SCREEN_SPACE_MIN_SELECTION_DISTANCE, 0)).subtract(window.getWorldSpacePoint(new Vector2(0, 0))).getMagnitude();
		currentStroke = new ArrayList<Point>();
		strokes = new ArrayList<ArrayList<Point>>();
		
		worldspaceCleanupThreshold = window.getWorldSpacePoint(new Vector2(CLEANUP_CONTROL_POINT_THRESHOLD, 0)).getMagnitude();
		
		
		window.setModuleGlassPane(pane);
		window.setModuleMouseListener(this);
		window.setModuleMouseMotionListener(this);
		window.setModuleKeyListener(this);
		
		poseListener = new PoseModeMouseListener(window);
	}
	
	private JToolBar makeToolBar() {
		JToolBar toolBar = new JToolBar(JToolBar.VERTICAL);

		JButton clearButton = new JButton();
		clearButton.setActionCommand("clear");
		clearButton.addActionListener(this);
		clearButton.setText("Clear");
		clearButton.setName("clear");
		toolBar.add(clearButton);
		
		JButton processButton = new JButton();
		processButton.setActionCommand("process");
		processButton.addActionListener(this);
		processButton.setText("Process");
		processButton.setName("process");
		toolBar.add(processButton);
		
		JButton selectAgentButton = new JButton();
		selectAgentButton.setActionCommand("selectAgent");
		selectAgentButton.addActionListener(this);
		selectAgentButton.setText("Select Agent");
		selectAgentButton.setName("selectAgent");
		toolBar.add(selectAgentButton);
		
		showMeButton = new JButton();
		showMeButton.setActionCommand("showMe");
		showMeButton.addActionListener(this);
		showMeButton.setText("What are you doing?!");
		showMeButton.setName("ShowMe");
		toolBar.add(showMeButton);
		
		return toolBar;
	}

	@Override
	public void deactivate() {
	}

	@Override
	public String getName() {
		return "Sketcho";
	}

	private void clearStatements() {
		;
		rootStatement = null;
		activeStatement = null;
		needsMoreStatements = true;
		intermediateStatements.clear();
		strokes.clear();
	}
	
	@Override
	public void actionPerformed(ActionEvent arg0) {
		if(arg0.getActionCommand().equals("clear")) {
			clearStatements();
		} else if(arg0.getActionCommand().equals("process")) {
			if(rootStatement == null)
				return;
			
			DefaultTreeNode<Statement> root = this.convertTree(rootStatement);
			//printTree(root, 0);
			clearStatements();
			parent.rawInputGathered(root);
		} else if (arg0.getActionCommand().equals("selectAgent")) {
			parent.activateAgentSelection();
		} else if (arg0.getActionCommand().equals("showMe")) {
			renderingStatements = !renderingStatements;
			
			if (!renderingStatements)
			{
				;
				renderedStatements.clear();
				
				showMeButton.setText("What are you doing?!");
			}
			else
				showMeButton.setText("Nevermind!");
		} else {
			selectNewStatement(arg0.getActionCommand());
		}
	}
	
	private Point[][] convertStrokes(ArrayList<ArrayList<Point>> stroke) {
		Point[][] pts = new Point[stroke.size()][];
		for(int i = 0; i < pts.length; i++) {
			pts[i] = new Point[stroke.get(i).size()];
			for(int j = 0; j < stroke.get(i).size(); j++) {
				pts[i][j] = stroke.get(i).get(j);
			}
		}
		return pts;
	}
	
	private void selectNewStatement(String name) {
		for(StatementName sn : StatementName.values()) {
			if(name.equals(sn.toString())) {
				return;
			}
		}
	}
	
	private WorldSpacePoint getIntendedWorldSpacePoint(Vector2 worldPoint) {
		SketchoStatement statement = getCollidedStatement(worldPoint);
		Poseable p = getIntendedPoseable(worldPoint);
		WorldSpacePoint toReturn;
		//Give preference to intermediate statements over poseables
		if(statement != null && !statement.fitsSyntaxTree() && statement != selectedStatement) {
			toReturn = new WorldSpacePointSketchListener(statement, worldPoint);
		} else if(p != null) {
			toReturn = new WorldSpacePointPoseable(p);
		} else {
			toReturn = new WorldSpacePointVector(worldPoint);
		}
		return toReturn;
	}
	
	private Poseable getIntendedPoseable(Vector2 worldPoint) {
		List<Poseable> poseables = window.getPoseablesInRange(worldPoint, worldspaceCleanupThreshold); 
		if(poseables.size() == 1) {
			return poseables.get(0);
		} else if(poseables.size() > 1) {
			return poseables.get(0);
		}
		else {		
			return null;
		}
	}
	
	/**
	 * Returns a statement if it is close enough to the world point to collide with it
	 * @param worldPoint
	 * @return
	 */
	private SketchoStatement getCollidedStatement(Vector2 worldPoint) {
		return getSelectedStatement(worldPoint, worldspaceCleanupThreshold * 1.5f, false);
	}
	
	public void setNewActiveWord(Vector2 worldPoint, String statementType) throws Exception {
		SketchoStatement newStatement;
		
		newStatement = factory.getStatement(statementType);
		
		newStatement.setTranslator(new ScreenSpaceWorldSpaceTranslator(window.getGame()));
		
		WorldSpacePoint wsp = getIntendedWorldSpacePoint(worldPoint);
		newStatement.setControlPoint1(wsp);
		
		activeStatement = new DefaultTreeNode<SketchoStatement>(newStatement);
	}
	
	public void setActiveStatementControlPoint2(Vector2 worldPoint) {
		if (activeStatement != null) {
			
			WorldSpacePoint wsp = getIntendedWorldSpacePoint(worldPoint);
			((SketchoStatement)activeStatement.getUserObject()).setControlPoint2(wsp);
		}
	}
	
	public void storeActiveStatement() {
		((SketchoStatement)activeStatement.getUserObject()).setId(new SketchoId());
		
		if(((SketchoStatement)activeStatement.getUserObject()).fitsSyntaxTree()) {
			generalizeStatement((SketchoStatement)activeStatement.getUserObject());
			
			
			if(currentParent != null) {
				try {
					activeStatement.setParent(currentParent);
					currentParent.addChild(activeStatement);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
			
			if(rootStatement == null) {
				rootStatement = activeStatement;
			}
			
			setupNextStatementParameters();
		} else {
			intermediateStatements.add((SketchoStatement)activeStatement.getUserObject());
		}
	}
	
	public void generalizeStatement(SketchoStatement statement)
	{
		Iterable<WorldSpacePoint> generalizableControlPoints = 
			statement.getGeneralizableControlPoints();
		
		int controlPointIndex = 0;
		
		for (WorldSpacePoint controlPoint : generalizableControlPoints)
		{
			if (controlPoint.holdsPoseable())
			{
				Poseable controlPoseable = controlPoint.getPoseable(); 
				
				Attribute generalization = generalizer.generalize(controlPoseable);
				
				if (generalization != null)
				{
					WorldSpacePoint generalizedControlPoint = 
						new WorldSpacePointAttribute(generalization);
					
					statement.setGeneralizableControlPoint(generalizedControlPoint, 
							controlPointIndex);
				}
			}
			
			controlPointIndex++;
		}
	}
	
	private void setupNextStatementParameters() {
		if(activeStatement != null && activeStatement.getChildCount() < ((SketchoStatement)activeStatement.getUserObject()).numberChildrenNeeded()) {  //latest statement needs children
			currentParent = activeStatement;
			activeStatement = null;
			needsMoreStatements = true;
		} else if (currentParent != null) {	//latest statement does not need children and is not root node
			currentParent = activeStatement.getParent();
			//activeStatement = currentParent;
			activeStatement = activeStatement.getParent();
			setupNextStatementParameters();
			
		} else { //latest statement does not need children and is root node. We are done
			activeStatement = null;
			needsMoreStatements = false;
		}
	}
	
	private ArrayList<Statement> getTreeStatementsAsList() {
		ArrayList<Statement> statements = new ArrayList<Statement>();
		if(rootStatement != null) {			
			treeToList(rootStatement, statements);
		}
		return statements;
	}
	
	@SuppressWarnings("unchecked")
	private void treeToList(DefaultTreeNode<SketchoStatement> base, ArrayList<Statement> list) {
		list.add(base.getUserObject());
		Iterator<? extends TreeNode> iterator = base.getChildren();
		while(iterator.hasNext()) {
			treeToList((DefaultTreeNode<SketchoStatement>) iterator.next(), list);
		}
	}
	
	@SuppressWarnings("unchecked")
	private DefaultTreeNode<Statement> convertTree(DefaultTreeNode<SketchoStatement> node) {
		DefaultTreeNode<Statement> root = new DefaultTreeNode<Statement>(node.getUserObject());
		Iterator<? extends TreeNode> iterator = node.getChildren();
		while(iterator.hasNext()) {
			root.addChild(convertTree((DefaultTreeNode<SketchoStatement>)iterator.next()));
		}
		return root;
	}
	
	private void setPaneLines(boolean includeActiveStatement) {
		ArrayList<Line> lines = new ArrayList<Line>();
		
		//printTree(rootStatement, 0);
		
		if(currentStroke != null && currentStroke.size() > 0) {
		   Point lastPoint = currentStroke.get(0);
		   for(Point p : currentStroke) {
			   lines.add(new Line(new Vector2((float)lastPoint.X, (float)lastPoint.Y),
					   new Vector2((float)p.X, (float)p.Y), SketchoStatement.UNSELECTION_COLOR));
			   lastPoint = p;
		   }
		}
		
		for(int i = 0; i < strokes.size(); i++) {
			ArrayList<Point> stroke = strokes.get(i);
			Point lastPoint = stroke.get(0);
			for(Point p : stroke) {
				lines.add(new Line(new Vector2((float)lastPoint.X, (float)lastPoint.Y), 
						new Vector2((float)p.X, (float)p.Y), SketchoStatement.UNSELECTION_COLOR));
				lastPoint = p;
			}
		}
		
		ArrayList<SketchoStatement> allStatements = new ArrayList<SketchoStatement>();
		//allStatements.addAll((Collection<? extends SketchoStatement>) getTreeStatementsAsList());
		allStatements.addAll(renderedStatements);
		allStatements.addAll(intermediateStatements);
		
		if(rootStatement != null)
			generateTreeLines(rootStatement, lines, LineFonts.solid);
		
		
		for(SketchoStatement sw : allStatements) {
			for(Line l : sw.getDrawLines()) {
				Vector2 p1, p2;
				p1 = window.getScreenSpacePoint(l.getPoint1());
				p2 = window.getScreenSpacePoint(l.getPoint2());
				lines.add(new Line(p1, p2, l.getColor()));
			}
		}
		
		
		if(includeActiveStatement && activeStatement != null && ((SketchoStatement)activeStatement.getUserObject()).isComplete()) {
			for(Line l : ((SketchoStatement)activeStatement.getUserObject()).getDrawLines()) {
				Vector2 p1, p2;
				p1 = window.getScreenSpacePoint(l.getPoint1());
				p2 = window.getScreenSpacePoint(l.getPoint2());
				lines.add(new Line(p1, p2, l.getColor()));
			}
		}
		
		pane.setLines(lines);
	}
	
	@SuppressWarnings("unchecked")
	public void generateTreeLines(DefaultTreeNode<SketchoStatement> node, ArrayList<Line> lines, LineFonts parentFont) {
		for(Line l : node.getUserObject().getDrawLines()) {
			Vector2 p1, p2;
			p1 = window.getScreenSpacePoint(l.getPoint1());
			p2 = window.getScreenSpacePoint(l.getPoint2());
			LineFonts thisFont;
			if(node.getChildCount() == 0 && node.getChildCount() == node.getUserObject().numberChildrenNeeded())
				thisFont = parentFont;
			else
				thisFont = l.getFont();
			lines.add(new Line(p1, p2, l.getColor(), thisFont));
		}
		
		Iterator<? extends TreeNode> iterator = node.getChildren();
		int i = 0;
		while(iterator.hasNext()) {
			LineFonts nextFont = node.getUserObject().getChildFont(i);
			generateTreeLines((DefaultTreeNode<SketchoStatement>)iterator.next(), lines, nextFont);
			i++;
		}
	}
	
	private void deselect() {
		if(selectedStatement != null)
			selectedStatement.deSelect();
		selectedStatement = null;
	}
	
	@SuppressWarnings("unchecked")
	private void eraseStatement(Statement s) {
		if(((SketchoStatement)s).fitsSyntaxTree()) {
			 DefaultTreeNode<SketchoStatement> node = getStatementNode(s, rootStatement);
			 if(node.getChildCount() > 0) {
				 List<String> answers = new ArrayList<String>();
				 String answer1 = "Delete the statement and its children";
				 String answer2 = "Replace the statement with its first child";
				 String answer3 = "Replace the statement with its second child";
				 
				 answers.add(answer1);
				 answers.add(answer2);	
				 
				 if(node.getChildCount() > 1) {				 
					 answers.add(answer3);
				 }
				 
				 String decision = "";
				try {
					decision = window.askMultipleChoiceQuestion("Do you want me to: ", answers);
				} catch (Exception e) {
					return; //Assume user cancelled
				}
				 if(decision.equals(answer1)) {
					 if(node.getParent() != null) {
						node.getParent().removeChild(node);
					 } else {
						 rootStatement = null;
					 }
				 } else if (decision.equals(answer2)) {
					 DefaultTreeNode<SketchoStatement> replacement = (DefaultTreeNode<SketchoStatement>) node.getOrderedChildren().get(0);
					 if(node.getParent() != null) {
						 DefaultTreeNode<SketchoStatement> parent = node.getParent();
						 parent.removeChild(node);
						 parent.addChild(replacement);
					 } else {
						 rootStatement = replacement;
					 }
				 } else {
					 DefaultTreeNode<SketchoStatement> replacement = (DefaultTreeNode<SketchoStatement>) node.getOrderedChildren().get(1);
					 if(node.getParent() != null) {
						 DefaultTreeNode<SketchoStatement> parent = node.getParent();
						 parent.removeChild(node);
						 parent.addChild(replacement);
					 } else {
						 rootStatement = replacement;
					 }
				 }
			 } else {
				 if(node.getParent() != null) {
					node.getParent().removeChild(node);
				 } else {
					 rootStatement = null;
				 }
			 }
			 if(!needsMoreStatements)
				 needsMoreStatements = true;
		} else {
			intermediateStatements.remove((SketchoStatement)s);
		}
		
		 if(activeStatement != null && ((SketchoStatement)activeStatement.getUserObject()).equals(s))
				activeStatement = null;
	}
	private DefaultTreeNode<SketchoStatement> getStatementNode(Statement s, DefaultTreeNode<SketchoStatement> defaultTreeNode) {
		if(defaultTreeNode.getUserObject().equals(s))
			return defaultTreeNode;
		else if(defaultTreeNode.getChildCount() == 0)
			return null;
		else {
			Iterator<? extends TreeNode> iterator = defaultTreeNode.getChildren();
			while(iterator.hasNext()) {
				@SuppressWarnings("unchecked")
				DefaultTreeNode<SketchoStatement> relevantChild = getStatementNode(s, (DefaultTreeNode<SketchoStatement>) iterator.next());
				if(relevantChild != null)
					return relevantChild;
			}
			return null;
		}
	}

	@Override
	public void mouseClicked(MouseEvent e) {
		buttonDown = e.getButton();
		if (buttonDown != MouseEvent.BUTTON1)
			poseListener.mouseClicked(e);
	}

	@Override
	public void mouseEntered(MouseEvent e) {
		if (e.getButton() == MouseEvent.BUTTON1)
			lastClick = window.getWorldSpacePoint(new Vector2(e.getX(), e.getY()));
		else
			poseListener.mouseEntered(e);
	}

	@Override
	public void mouseExited(MouseEvent e) {
		if (e.getButton() != MouseEvent.BUTTON1)
			poseListener.mouseExited(e);
	}
	
	private SketchoStatement getSelectedStatement(Vector2 worldPos, float maxDist, boolean includeSelectedStatement) {
		float minDistance = Float.MAX_VALUE;
		SketchoStatement closest = null;
		
		if(selectedStatement != null && includeSelectedStatement) {
			float dist = selectedStatement.distanceToPoint(worldPos);
			if(dist < maxDist)
				return selectedStatement;
		}
		
		for(SketchoStatement ss : intermediateStatements) {
			float dist = ss.distanceToPoint(worldPos);
			if(dist < minDistance) {
				closest = ss;
				minDistance = dist;
			}
		}
		
		if(minDistance < maxDist)
			return closest;
		
		for(Statement s : getTreeStatementsAsList()) {
			SketchoStatement ss = (SketchoStatement)s;
			float dist = ss.distanceToPoint(worldPos);
			if (dist < minDistance) {
				closest = ss;
				minDistance = dist;
			}
		}
		
		if(minDistance < maxDist)
			return closest;
		else
			return null;
	}
	
	@SuppressWarnings({ "unused", "unchecked" })
	private void printTree(DefaultTreeNode node, int numSpaces) {
		if(node == null) {
			System.out.println("Null node, returning");
			return;
		}
		
		for(int i = 0; i < numSpaces; i++)
			System.out.print(" ");
		System.out.print("|");
		System.out.println(node.getUserObject());
		
		Iterator<? extends TreeNode> iterator = node.getChildren();
		while(iterator.hasNext()) {
			printTree((DefaultTreeNode) iterator.next(), numSpaces + 1);
		}
	}
	
	private void insertStatement(Result result) {
		Vector2 controlPoint1 = window.getWorldSpacePoint(new Vector2(result.getControlPoints()[0]));
		try {
			setNewActiveWord(controlPoint1, result.getName());
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		Vector2 controlPoint2 = window.getWorldSpacePoint(new Vector2(result.getControlPoints()[1]));
		setActiveStatementControlPoint2(controlPoint2);
		storeActiveStatement();
		strokes.clear();
	}
	
	@Override
	public void mousePressed(MouseEvent e) {
		buttonDown = e.getButton();
		if (buttonDown != MouseEvent.BUTTON1)
		{
			poseListener.mousePressed(e);
			
			return;
		}
		
		Vector2 click = window.getWorldSpacePoint(new Vector2(e.getX(), e.getY()));
		
		SketchoStatement selection = getSelectedStatement(click, minDistanceToSelect, true);
		deselect();
		if(selection  != null && strokes.size() == 0) {  //We don't select statements if we're mid drawing
			selection.select(click);
			selectedStatement = selection;
		} else {
			currentStroke.add(new Point(e.getPoint()));
			selectedStatement = null;
		}
		
		
		recognitionTime = -1;

		lastClick = click;		
	}
	
	private void repaintPane(boolean includeActiveStatement)
	{
		setPaneLines(includeActiveStatement);
		pane.repaint();
		
	}

	@Override
	public void mouseReleased(MouseEvent e) {
		buttonDown = 0;
		if (e.getButton() != MouseEvent.BUTTON1)
		{
			poseListener.mouseReleased(e);
			
			return;
		}
		
		if(currentStroke.size() > 1)
			strokes.add(currentStroke);	
		
		currentStroke = new ArrayList<Point>();
		
		if(strokes.size() > 0) {
			ArrayList<Result> results = recognizer.recognize(convertStrokes(strokes), true, false);
			if(results.size() == 0) {
				recognitionTime = System.currentTimeMillis() + MORE_INPUT_WAIT_TIME;
				
				return;
			}
			Result result = results.get(0);
			boolean ambiguous = false;
			if(results.size() > 1 && result.getScore() - results.get(1).getScore() < MULTI_RESULT_AMBIGUITY_THRESHOLD)
				ambiguous = true;
			System.out.println("Results: " + results);
			
			if(result.getScore() > FULL_CONFIDENCE_THRESHOLD && !ambiguous) {
				insertStatement(result);
				strokes.clear();
			} else if(ambiguous) {
				recognitionTime = System.currentTimeMillis() + AMBIGUOUS_RECOGNITION_STALL_TIME;
			} else  {
				recognitionTime = System.currentTimeMillis() + MORE_INPUT_WAIT_TIME;
			}
		}
		
		
	}

	@Override
	public void mouseDragged(MouseEvent arg0) {
		if (buttonDown != MouseEvent.BUTTON1)
		{
			poseListener.mouseDragged(arg0);
			
			return;
		}
		
		Vector2 click = window.getWorldSpacePoint(new Vector2(arg0.getX(), arg0.getY()));
		WorldSpacePoint wsp = getIntendedWorldSpacePoint(click);
		if (selectedStatement != null) {
			selectedStatement.mouseDrag(click.subtract(lastClick), wsp);
		} else {
			currentStroke.add(new Point(arg0.getPoint()));
		}
		
		recognitionTime = -1;
		lastClick = click;
	}

	@Override
	public void mouseMoved(MouseEvent arg0) {
		if (arg0.getButton() != MouseEvent.BUTTON1)
			poseListener.mouseMoved(arg0);
	}
	
	public void update() {
	    	pane.setLineWidth(5);
		
		if(recognitionTime != -1) {
			long timeNow = System.currentTimeMillis();
			if(timeNow >= recognitionTime && strokes.size() > 0) {
				ArrayList<Result> results = recognizer.recognize(convertStrokes(strokes), true, false);
				if(results.size() == 0) {
					strokes.clear(); 
				} else {
					Result result = results.get(0);
					if(result.getScore() > MINIMUM_RECOGNITION_THRESHOLD)
						insertStatement(result);
					else
						strokes.clear();
				}
				recognitionTime = -1;
			}
			
		}
		
		if (renderingStatements)
			refreshRenderedStatements();
		
		if(activeStatement == null || !((SketchoStatement)activeStatement.getUserObject()).isComplete())
			repaintPane(false);
		else
			repaintPane(true);	
	}

	@Override
	public void tick() {
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				update();
			}
		});
			
	}

	@Override
	public void keyPressed(KeyEvent arg0) {
		if(arg0.getKeyCode() == KeyEvent.VK_DELETE && selectedStatement != null) {
			eraseStatement(selectedStatement);
			selectedStatement = null;						
		}
		
	}

	@Override
	public void keyReleased(KeyEvent arg0) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void keyTyped(KeyEvent arg0) {
		// TODO Auto-generated method stub
		
	}
	
	private void refreshRenderedStatements()
	{
		List<SketchoStatement> statementsToRender = new LinkedList<SketchoStatement>();
		
		AIController agent = parent.getAIAgent().getAIController();
		WorldState currentState = window.getGame().getCurrentWorldState();
		//FIXME: IBT Reliant. Should be AI Independant. List of Nodes, maybe?
		IBTNode[] activePath = agent.getActivePath(currentState);
		
		/*
		for (int i = 0; i < activePath.length - 1; i++)
		{
			//FIXME: NO IBT REFERENCES!
			IBTDecisionNode decisionNode = (IBTDecisionNode) activePath[i];
			Attribute decision = decisionNode.getAttribute();
			SketchoStatement decisionStatement = factory.getStatement(decision);
			statementsToRender.add(decisionStatement);
		}
		*/
		
		Action currentAction = ((IBTLeafNode) activePath[activePath.length - 1]).getAction();
		SketchoStatement actionStatement = factory.getStatement(currentAction);
		statementsToRender.add(actionStatement);
		
		renderedStatements.clear();
		for (SketchoStatement statementToRender : statementsToRender)
		{
			if (statementToRender != null)
			{
				statementToRender.setTranslator(
						new ScreenSpaceWorldSpaceTranslator(window.getGame()));
				renderedStatements.add(statementToRender);
			}
		}
	}
}